//
// C++ Implementation: menu
//
// Description:
//
//
// Author: Dave Parker <drp@ezeedsl.co.uk >, (C) 2006
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "font.h"
#include "menu.h"
#include "surface.h"

using namespace std;

/**
 * Create an emtpy menu with a given title and subtitle
 * @param t - Optional title
 * @param s - Optional subtitle
 * @return
 */
Menu::Menu ( const std::string& t, const std::string& s )
{
    max_len = 0;
    current_option = 0;
    title = t;
    subtitle = s;
    callback = NULL;
    callback_data = NULL;
}


/**
 * Destructor
 * @return
 */
Menu::~Menu()
{
}



/**
 * Add an option to the menu
 * @param option
 */
void Menu::Add ( const std::string& option )
{
    options.push_back ( option );
    if ( option.length() > max_len )
    {
        max_len = option.length();
    }
}



/**
 * Render menu to Surface
 * @param s - Destination Surface
 * @param f - Font
 * @param fg - Foreground Colour
 * @param bg - Background Colour
 * @param hl - Highlight Colour for selected option
 */
void Menu::Render ( Surface &s, const Font& f, const Colour &fg, const Colour &bg, const Colour &hl )
{
    int spacing = f.GetSize() * 2.5;

    double n=0;

    if ( title != "" )
    {
        Centre ( s, f, fg, bg, title, f.GetSize(), 0 );
        n++;
    }

    if ( subtitle != "" )
    {
        Centre ( s, f, fg, bg, subtitle, f.GetSize() + spacing, 0 );

        if (options.size() < 5)
        {
            n++;
        }
        else
        {
            n += 0.4;
        }
    }

    for ( int i=0; i<options.size(); i++ )
    {
        Colour col ( bg );
        if ( i == current_option )
        {
            col = hl;
        }
        Centre ( s, f, fg, col, options[i],  f.GetSize() +( i+n+1 ) *spacing, max_len );
    }

    if ( caption != "" )
    {
        Centre ( s, f, fg, bg, caption, 240 -f.GetSize() *2, 0 );
    }

}



/**
 * Move up the menu
 */
void Menu::Up()
{
    current_option = ( current_option + options.size() - 1 ) % options.size();
}


/**
 * Move down the menu
 */
void Menu::Down()
{
    current_option = ( current_option + 1 ) % options.size();
}



/**
 * Returns the selected option
 * @return Selected option, first is 0
 */
int Menu::GetOption()
{
    return current_option;
}



/**
 * Set the selection callback
 * @param (* fn)( void * ) - Callback Function
 * @param data - user defined data passed to the calback function
 */
void Menu::SetCallback ( void ( *fn ) ( void* ), void* data )
{
    callback = fn;
    callback_data = data;
}


/**
 * Call callback if set
 */
void Menu::DoAction()
{
    if ( callback != NULL )
    {
        ( *callback ) ( callback_data );
    }
}



/**
 * Display a centred a text string in a box
 * @param s - Destination Surface
 * @param f - Font
 * @param fg - Foreground Colour
 * @param bg - Background Colour
 * @param text
 * @param y - Y position
 * @param len - String length in characters
 */
void Menu::Centre ( Surface &s, const Font& f, const Colour &fg, const Colour &bg,
                    const std::string& text, int y, int len )
{
    len = len * f.GetSize();

    int w,h;
    f.GetTextSize ( text, w, h );
    if ( len == 0 )
    {
        len = w + f.GetSize() * 2;
    }

    int x = ( s.GetWidth() - len ) /2;

    s.Rect ( x, y - h/2, len, h * 2, bg );

    f.PutString ( s, ( s.GetWidth() - w ) /2, y, fg, bg, text );
}



/**
 * Set the selected option
 * @param opt
 */
void Menu::SetOption ( int opt )
{
    current_option = opt % options.size();
}



/**
 * Set the caption at the bottom of the menu
 * @param s - Caption
 */
void Menu::SetCaption ( const std::string& s )
{
    caption = s;
}


/**
 * Set the subtitle of the menu
 * @param s - Subtitle
 */
void Menu::SetSubtitle ( const std::string& s )
{
    subtitle = s;
}
