/***************************************************************************
 *   Copyright (C) 2006 by Dave Parker                                     *
 *   drp@ezeedsl.co.uk                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "board.h"
#include "piece.h"

using namespace std;

/**
 * Construct an empty board
 * @return
 */
Board::Board()
{
    for ( int i=0; i<8; i++ )
    {
        for ( int j=0; j<8; j++ )
        {
            board[i][j] = NULL;
        }
    }
}


/**
 * Copy constructor
 * @param b - Source board
 * @return
 */
Board::Board ( const Board& b )
{
    for ( int i=0; i<8; i++ )
    {
        for ( int j=0; j<8; j++ )
        {
            if ( b.board[i][j] == NULL )
            {
                board[i][j] = NULL;
            }
            else
            {
                board[i][j] = b.board[i][j]->Clone();
            }
        }
    }

    for ( list<Piece*>::const_iterator it = b.taken.begin(); it != b.taken.end(); it++ )
    {
        taken.push_back ( ( *it )->Clone() );
    }
}


/**
 * Destructor
 * @return
 */
Board::~Board()
{
    for ( int i=0; i<8; i++ )
    {
        for ( int j=0; j<8; j++ )
        {
            if ( board[i][j] != NULL )
            {
                delete board[i][j];
            }
        }
    }
}


/**
 * Return piece at a given position
 * @param x
 * @param y
 * @return
 */
Piece* Board::Get ( int x, int y )
{
    return board[x][y];
}



/**
 * Move a piece
 * @param x1
 * @param y1
 * @param x2
 * @param y2
 * @param promote - Piece to which to promote
 * @param sn - Skin name
 */
void Board::Move ( int x1, int y1, int x2, int y2, char promote, const std::string& sn )
{
    Side s = board[x1][y1]->GetSide();

    board[x1][y1]->Move ( *this, x1, y1, x2, y2 );

    switch ( promote )
    {
    case 'q':
        delete board[x2][y2];
        board[x2][y2] = new Queen ( s, sn + "/queen" );
        break;

    case 'r':
        delete board[x2][y2];
        board[x2][y2] = new Rook ( s, sn + "/rook" );
        break;

    case 'n':
        delete board[x2][y2];
        board[x2][y2] = new Knight ( s, sn + "/knight" );
        break;

    case 'b':
        delete board[x2][y2];
        board[x2][y2] = new Bishop ( s, sn + "/bishop" );
        break;
    }
}


/**
 * Put a piece at a given position
 * @param x
 * @param y
 * @param p - Piece
 */
void Board::Set ( int x, int y, Piece* p )
{
    board[x][y] = p;
}



/**
 * Set up board in start position
 * @param sn - Skin name
 */
void Board::Setup ( const std::string& sn )
{
    for ( int i=0; i<8; i++ )
    {
        for ( int j=2; j<6; j++ )
        {
            if ( board[i][j] != NULL )
            {
                delete board[i][j];
            }
            board[i][j] = NULL;
        }
        board[i][1] = new Pawn ( BLACK, sn + "/pawn" );
        board[i][6] = new Pawn ( WHITE, sn + "/pawn" );
    }

    board[0][0] = new Rook ( BLACK, sn + "/rook" );
    board[7][0] = new Rook ( BLACK, sn + "/rook" );
    board[1][0] = new Knight ( BLACK, sn + "/knight" );
    board[6][0] = new Knight ( BLACK, sn + "/knight" );
    board[2][0] = new Bishop ( BLACK, sn + "/bishop" );
    board[5][0] = new Bishop ( BLACK, sn + "/bishop" );
    board[3][0] = new Queen ( BLACK, sn + "/queen" );
    board[4][0] = new King ( BLACK, sn + "/king" );

    board[0][7] = new Rook ( WHITE, sn + "/rook" );
    board[7][7] = new Rook ( WHITE, sn + "/rook" );
    board[1][7] = new Knight ( WHITE, sn + "/knight" );
    board[6][7] = new Knight ( WHITE, sn + "/knight" );
    board[2][7] = new Bishop ( WHITE, sn + "/bishop" );
    board[5][7] = new Bishop ( WHITE, sn + "/bishop" );
    board[3][7] = new Queen ( WHITE, sn + "/queen" );
    board[4][7] = new King ( WHITE, sn + "/king" );

    ClearTaken();
}



/**
 * Assignment operator
 * @param b - Source board
 * @return
 */
const Board& Board::operator= ( const Board& b )
{
    // Clone pieces
    for ( int i=0; i<8; i++ )
    {
        for ( int j=0; j<8; j++ )
        {
            Piece* p = NULL;
            if ( b.board[i][j] != NULL )
            {
                p = b.board[i][j]->Clone();
            }

            if ( board[i][j] != NULL )
            {
                delete board[i][j];
            }
            board[i][j] = p;
        }
    }

    // Clone take list
    ClearTaken();
    for ( list<Piece*>::const_iterator it = b.taken.begin(); it != b.taken.end(); it++ )
    {
        taken.push_back ( ( *it )->Clone() );
    }

    return *this;
}

/**
 * Clear the taken piece list
 */
void Board::ClearTaken()
{
    while ( !taken.empty() )
    {
        delete taken.front();
        taken.pop_front();
    }
}

/**
 * Take a piece at a given position
 * @param x
 * @param y
 */
void Board::Take ( int x, int y )
{
    // Cancel out types for different sides
    Piece* p = board[x][y];
    int n = 0;
    bool found = false;
    for ( list<Piece*>::iterator it = taken.begin(); !found && it != taken.end(); it++ )
    {
        if ( ( *it )->GetType() == p->GetType() && ( *it )->GetSide() != p->GetSide() )
        {
            found = true;
            delete ( *it );
            taken.erase ( it );
            delete p;
        }
    }

    // If not in the list, add it
    if ( !found )
    {
        taken.push_back ( board[x][y] );
    }

    board[x][y] = NULL;
}



/**
 * Find difference in number of taken pieces
 * @param t - Type of piece
 * @param n - Returned count
 * @return Pointer to the piece, NULL if not found
 */
Piece* Board::Taken ( PieceType t, int& n )
{
    Piece* p = NULL;
    n = 0;
    for ( list<Piece*>::iterator it = taken.begin(); it != taken.end(); it++ )
    {
        if ( ( *it )->GetType() == t )
        {
            p = *it;
            n++;
        }
    }

    return p;
}
