/***************************************************************************
 *   Copyright (C) 2006 by Dave Parker                                     *
 *   drp@ezeedsl.co.uk                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "chess2x.h"
#include "colour.h"
#include "font.h"
#include "image.h"
#include "menu.h"
#include <algorithm>
#include <dirent.h>
#include <iomanip>
#include <fstream>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

using namespace std;

/**
 * Initialise application
 * @param s - Screen
 */
void Chess2X::Initialise ( Screen& s )
{
    config.LoadFile ( "options" );
    GetSkins();
    SetSkin();

    font = new Font ( "font", 11 );
    GetSounds();
    GetVolume();

    board_x = board_y = 0;
    inpipefd[0] = inpipefd[1] = -1;
    outpipefd[0] = outpipefd[1] = -1;
    pid = -1;
    move_list_size = 4;
    mode = MENU;
    old_mode = MENU;
    game_mode = ONE_PLAYER_WHITE;
    hash_size = 8192;
    config.GetValue ( "hashsize", hash_size );
    config.SetValue ( "hashsize", hash_size );

    main_menu = new Menu ( "Chess2X" );
    main_menu->Add ( "Play" );
    main_menu->Add ( "Load Game" );
    main_menu->Add ( "Options" );

    main_menu->SetCaption ( "B - Select  Start - Exit  A - Player" );
    main_menu->SetCallback ( MainAction, this );

    start_menu = new Menu ( "Chess2X" );
    start_menu->Add ( "Play white" );
    start_menu->Add ( "Play black" );
    start_menu->Add ( "Two players" );
    start_menu->SetCaption ( "B - Select  X - Main menu  Start - Exit  A - Player" );
    start_menu->SetCallback ( StartAction, this );

    in_game_menu = new Menu ( "Chess2X" );
    in_game_menu->Add ( "Return to game" );
     in_game_menu->Add ( "Hint" );
    in_game_menu->Add ( "Take back move" );    in_game_menu->Add ( "Save" );
    in_game_menu->Add ( "Abandon game" );
	in_game_menu->SetCaption ( "B - Select  Start - Exit  A - Player" );
    in_game_menu->SetCallback ( InGameAction, this );

    quit_menu = new Menu ( "Chess2X", "Are you sure you want to quit?" );
    quit_menu->Add ( "Yes" );
    quit_menu->Add ( "No" );
    quit_menu->SetCallback ( QuitAction, this );

    promote_menu = new Menu ( "Promote pawn to" );
    promote_menu->Add ( "Queen" );
    promote_menu->Add ( "Rook" );
    promote_menu->Add ( "Knight" );
    promote_menu->Add ( "Bishop" );
    promote_menu->SetCaption ( "B - Select" );
    promote_menu->SetCallback ( PromoteAction, this );

    level_menu = new Menu ( "Chess2X" );
    level_menu->Add ( "Easier" );
    level_menu->Add ( "Harder" );
    level_menu->SetCaption ( "B - Select  X - Return to main menu  Start - Exit" );
    level_menu->SetCallback ( LevelAction, this );

    moves_per_period = 60;
    config.GetValue ( "moves_per_period", moves_per_period );
    config.SetValue ( "moves_per_period", moves_per_period );
    ostringstream oss;
    oss << "Level " << 11 - moves_per_period/10;
    level_menu->SetSubtitle ( oss.str() );

    options_menu = new Menu ( "Chess2X" );
    options_menu->Add ( "Set Level" );
    options_menu->Add ( "Select Skin" );
    options_menu->SetCaption ( "B - Select  X - Return to main menu  Start - Exit" );
    options_menu->SetCallback ( OptionsAction, this );

    skin_menu = new Menu ( "Chess2X","default" );
    skin_menu->SetSubtitle ( *skin );
    skin_menu->Add ( "Next" );
    skin_menu->Add ( "Previous" );
    skin_menu->SetCaption ( "B - Select  X - Return to main menu  Start - Exit" );
    skin_menu->SetCallback ( SkinAction, this );

    current_menu = main_menu;

    selector.SetCallback ( AudioFileSelected, this );

    repeats.push_back(Input::DOWN);
    repeats.push_back(Input::DOWNLEFT );
    repeats.push_back(Input::DOWNRIGHT);
    repeats.push_back(Input::L);
    repeats.push_back(Input::LEFT);
    repeats.push_back(Input::R);
    repeats.push_back(Input::UP);
    repeats.push_back(Input::UPLEFT);
    repeats.push_back(Input::UPRIGHT);
    repeats.push_back(Input::VOLDOWN);
    repeats.push_back(Input::VOLUP);

    Run();
}


/**
 * Render game/menus
 * @param s
 */
void Chess2X::Render ( Screen& s )
{
    switch ( mode )
    {
    case GAME_MENU:
    case MENU:
        DoMenu ( s );
        break;

    case GAME:
        Game ( s );
        break;

    case AUDIO_PLAYER:
        Player ( s );
        break;

    case FILE_SELECT:
        FileSelect ( s );
        break;

    }
}


/**
 * Handle button presses
 * @param input
 * @param b
 */
void Chess2X::ButtonChange ( const Input& input, Input::ButtonType b )
{
    if ( b == Input::START && input.IsPressed ( b ) )
    {
        old_mode = mode;
        mode = MENU;
        current_menu = quit_menu;
        current_menu->SetOption ( 1 );
    }
    else if ( b == Input::LEFT && input.IsPressed ( b ) )
    {
        LeftSquare();
    }
    else if ( b == Input::RIGHT && input.IsPressed ( b ) )
    {
        RightSquare();
    }
    else if ( b == Input::DOWN && input.IsPressed ( b ) )
    {
        if ( mode == GAME )
        {
            DownSquare();
        }
        else if ( mode == AUDIO_PLAYER )
        {
            player.NextFile();
        }
        else if ( mode == FILE_SELECT )
        {
            selector.NextFile();
        }
        else
        {
            current_menu->Down();
        }
    }
    else if ( b == Input::UP && input.IsPressed ( b ) )
    {
        if ( mode == GAME )
        {
            UpSquare();
        }
        else if ( mode == AUDIO_PLAYER )
        {
            player.PreviousFile();
        }
        else if ( mode == FILE_SELECT )
        {
            selector.PreviousFile();
        }
        else
        {
            current_menu->Up();
        }
    }
    else if ( b == Input::UPLEFT && input.IsPressed ( b ) )
    {
        UpSquare();
        LeftSquare();
    }
    else if ( b == Input::UPRIGHT && input.IsPressed ( b ) )
    {
        UpSquare();
        RightSquare();
    }
    else if ( b == Input::DOWNLEFT && input.IsPressed ( b ) )
    {
        DownSquare();
        LeftSquare();
    }
    else if ( b == Input::DOWNRIGHT && input.IsPressed ( b ) )
    {
        DownSquare();
        RightSquare();
    }
    else if ( b == Input::B && input.IsPressed ( b ) )
    {
        if ( mode == GAME )
        {
            if ( win )
            {
                mode = MENU;
                current_menu = main_menu;
                ResetGame();
            }
            else if ( ( whites_turn && game_mode == ONE_PLAYER_WHITE ) ||
                      ( !whites_turn && game_mode == ONE_PLAYER_BLACK ) ||
                      game_mode == TWO_PLAYER )
            {
                move += board_x + 'a';
                move += '0' + ( 8 - board_y );
                if ( move.length() == 4 )
                {
                    bool promote = false;
                    if ( board.Get ( select_x, select_y ) != NULL )
                    {
                        promote = board.Get ( select_x, select_y )->CheckPromote (
                                      select_x, select_y, board_x, board_y );
                    }

                    if ( promote )
                    {
                        current_menu = promote_menu;
                        promote_menu->SetOption ( 0 );
                        mode = GAME_MENU;
                    }
                    else
                    {
                        if ( move.substr ( 0,2 ) != move.substr ( 2,2 ) )
                        {
                            Send ( move+"\n" );
                            cout << "Move: " << move << endl;
                        }
                        move="";
                    }
                }
                select_x = board_x;
                select_y = board_y;
            }
        }
        else if ( mode == AUDIO_PLAYER )
        {
            player.ToggleState();
        }
        else if ( mode == FILE_SELECT )
        {
            selector.Select();
        }
        else
        {
            current_menu->DoAction();
        }
    }
    else if ( b == Input::SELECT && input.IsPressed ( b ) )
    {
        if ( mode == GAME )
        {
            old_mode = mode;
            mode = GAME_MENU;
            current_menu = in_game_menu;
            in_game_menu->SetOption ( 0 );
        }
    }
    else if ( b == Input::A && input.IsPressed ( b ) )
    {
        if ( mode == AUDIO_PLAYER )
        {
            mode = old_mode;
        }
        else if ( mode != FILE_SELECT )
        {
            old_mode = mode;
            mode = AUDIO_PLAYER;
        }
    }
    else if ( b == Input::L && input.IsPressed ( b ) )
    {
        if ( scroll > 0 ) scroll--;
    }
    else if ( b == Input::R && input.IsPressed ( b ) )
    {
        if ( ( scroll + move_list_size ) < output.size() ) scroll++;
    }
    else if ( b == Input::VOLUP && input.IsPressed ( b ) )
    {
        if ( mode == AUDIO_PLAYER )
        {
            player.VolumeUp();
            config.SetValue ( "musicvolume", player.GetVolume() );
        }
        else
        {
            move_sound->VolumeUp();
            take_sound->VolumeUp();
            config.SetValue ( "fxvolume", move_sound->GetVolume() );
			move_sound->Play();
        }
    }
    else if ( b == Input::VOLDOWN && input.IsPressed ( b ) )
    {
        if ( mode == AUDIO_PLAYER )
        {
            player.VolumeDown();
            config.SetValue ( "musicvolume", move_sound->GetVolume() );
        }
        else
        {
            move_sound->VolumeDown();
            take_sound->VolumeDown();
            config.SetValue ( "fxvolume", move_sound->GetVolume() );
 			move_sound->Play();
       }
    }
    else if ( b == Input::X && input.IsPressed ( b ) )
    {
        if ( mode == MENU )
        {
            current_menu = main_menu;
        }
        else if ( mode == AUDIO_PLAYER )
        {
            player.RemoveFile();
        }
        else if ( mode == FILE_SELECT )
        {
            selector.UpLevel();
        }
        else if ( mode == GAME )
        {
            if ( ( whites_turn && game_mode == ONE_PLAYER_BLACK ) ||
                    ( !whites_turn && game_mode == ONE_PLAYER_WHITE ) )
            {
                Message ( "Thinking","Forcing move" );
                kill ( pid, SIGINT );
            }
        }
    }
    else if ( b == Input::Y && input.IsPressed ( b ) )
    {
        if ( mode == AUDIO_PLAYER )
        {
            mode = FILE_SELECT;
            selector.SetMessage ( "" );
        }
        else if ( mode == FILE_SELECT )
        {
            mode = AUDIO_PLAYER;
        }
    }
}


/**
 * Run chess engine
 * @param
 */
void Chess2X::Run ( void )
{
    if ( pid != -1 )
    {
        kill ( pid, 9 );
        pid = -1;
        int status;
        wait ( &status );
    }



    if ( pipe ( inpipefd ) < 0 ) perror ( "inpipe" );
    if ( pipe ( outpipefd ) < 0 ) perror ( "outpipe" );

    pid = fork();

    if ( pid == 0 )
    {
        // Give priority to GUI
        nice ( 10 );

        // Child process - redirect stdout and stderr to outpipe
        close ( STDOUT_FILENO );
        dup2 ( outpipefd[1], STDOUT_FILENO );
        dup2 ( STDOUT_FILENO, STDERR_FILENO );

        // Close read end of pipe
        close ( outpipefd[0] );

        // Redirect stdin to inpipe
        close ( STDIN_FILENO );
        dup2 ( inpipefd[0], STDIN_FILENO );

        // Close write end of pipe
        close ( inpipefd[1] );

        // Remove buffering
        setbuf ( stdout, NULL );

        // Exec xgrep
        execl ( "gnuchess", "gnuchess", "-x", NULL );
        // execl("crafty", "crafty", "xboard", NULL);

        // Shouldn't get here
        perror ( "execl" );
        return;
    }
    else
    {
        // Parent process - close write end of outpipe
        close ( outpipefd[1] );
        // Parent process - close read end of inpipe
        close ( inpipefd[0] );
    }
}


/**
 * Read input from chess engine
 * @param input
 */
void Chess2X::NonRender ( const Input& input )
{
    char buf[1];
    int n;
    fd_set rfds, wfds, efds;

    // Check for I/O from gnuchess
    if ( outpipefd[0] != -1 )
    {
        FD_ZERO ( &rfds );
        FD_ZERO ( &wfds );
        FD_ZERO ( &efds );
        FD_SET ( outpipefd[0], &rfds );
        timeval timeout;
        timeout.tv_sec = 0;
        timeout.tv_usec = 100000;
        n = select ( outpipefd[0]+1, &rfds, &wfds, &efds, &timeout );
        if ( n == -1 )
        {
            perror ( "select" );
        }
        else if ( n > 0 && FD_ISSET ( outpipefd[0], &rfds ) )
        {
            // Output on pipe
            string line;
            bool line_read = false;
            while ( !line_read && ( n = read ( outpipefd[0], buf, 1 ) ) > 0 )
            {
                if ( buf[0] != '\n' )
                {
                    line += buf[0];
                }
                else
                {
                    Parse ( line );
                    line_read = true;
                }
            }

            if ( n <= 0 )
            {
                // Pipe closed - end of output
                close ( outpipefd[0] );
                outpipefd[0] = -1;
                pid=-1;

                close ( inpipefd[1] );
                inpipefd[1] = -1;

                // Clear up child process
                int wait_status = 0;
                wait ( &wait_status );
            }
        }
    }


    // Do repeat keys
    for (int i=0; i < repeats.size(); i++)
    {
        if ( input.Repeat ( repeats[i] ) )
        {
            ButtonChange ( input, repeats[i] );
        }
    }
}



/**
 * Send commands to the chess engine
 * @param s - String to send
 */
void Chess2X::Send ( const std::string& s )
{
    cout << "Send: " << s << flush;
    if ( s.length() > 0 )
    {
        write ( inpipefd[1], s.c_str(), s.length() );
    }
}



/**
 * Display a message
 * @param m - Line 1
 * @param m2 - Line 2
 * @param m3 - Line 3
 * @param m4 - Line 4
*/
void Chess2X::Message ( const std::string& m, const std::string& m2, const std::string& m3)
{
    message[0] = m;
    message[1] = m2;
    message[2] = m3;
}



/**
 * Parse output of chess engine
 * @param line
 */
void Chess2X::Parse ( const std::string line )
{
    cout << "Got " << line << endl;
    win = Win ( line );
    if ( !win && !Illegal ( line ) && !Hint ( line ) )
    {
        string::size_type pos = line.find ( '.' );
        if ( pos != string::npos )
        {
            int move_no = 0;
            istringstream iss ( line );
            iss >> move_no;
            if ( move_no > 0 )
            {
                cout << "Move: " << current_move << endl;
                int n = 2;
                if ( line[pos+2] == '.' )
                {
                    n = 6;
                }

                Message ( "Last move ", line.substr ( pos+n ), "Select - Menu");
                int x1 = line[pos+n] - 'a';
                int y1 = 8- ( line[pos+n+1] - '0' );
                int x2 = line[pos+n+2] - 'a';
                int y2 = 8- ( line[pos+n+3] - '0' );

                if ( board.Get ( x2, y2 ) == NULL )
                {
                    // Play move sound
                    move_sound->Play();
                }
                else
                {
                    // Play take sound
                    take_sound->Play();
                }

                board.Move ( x1, y1, x2, y2, line[pos+n+4], "skins/" + *skin );
                last_x = x2;
                last_y = y2;

                int l = 4;
                string promotion ( "qrnb" );
                if ( promotion.find ( line[pos+n+4] ) != string::npos )
                {
                    l++;
                }

                // Save history
                history.push_back ( History ( current_move, board, line.substr ( pos+n,l ),
                                              !whites_turn, last_x, last_y, board_x, board_y ) );

                if ( output.size() < current_move )
                {
                    if ( game_mode == ONE_PLAYER_WHITE )
                    {
                        Message ( "Thinking","","X - Force move" );
                        if ( move_queue.empty() && loaded )
                        {
                            Send ( "black\ngo\n" );
                            loaded = false;
                        }
                    }
                    whites_turn = false;
                    ostringstream oss;
                    oss << setw ( 2 ) << current_move << " " << line.substr ( pos+n,l ) << " ";
                    output.insert ( output.end(), oss.str() );
                    if ( current_move > move_list_size )
                    {
                        scroll = current_move - move_list_size;
                    }
                }
                else
                {
                    if ( game_mode == ONE_PLAYER_BLACK )
                    {
                        Message ( "Thinking","","X - Force move" );
                        if ( move_queue.empty() && loaded )
                        {
                            Send ( "white\ngo\n" );
                            loaded = false;
                        }
                    }
                    whites_turn = true;
                    output[current_move - 1] += line.substr ( pos+n,l );
                    current_move++;
                }

                // Do next queue move (load game)
                if ( !move_queue.empty() )
                {
                    Send ( move_queue.front() );
                    move_queue.pop_front();
                }
            }
        }
    }
}



/**
 * Check for win/draw message from chess engine
 * @param line
 * @return True if win/draw is detected
 */
bool Chess2X::Win ( const std::string& line )
{
    if ( line.substr ( 0, 3 ) == "1-0" )
    {
        Message ( "White wins", "B - continue" );
        return true;
    }

    if ( line.substr ( 0, 3 ) == "0-1" )
    {
        Message ( "Black wins", "B - continue" );
        return true;
    }

    if ( line.substr ( 0, 7 ) == "1/2-1/2" )
    {
        Message ( "Draw", "B - continue" );
        return true;
    }

    return false;
}



/**
 * Display game
 * @param s - Screen
 */
void Chess2X::Game ( Screen& s )
{
    Colour col1 ( s,255,0,0 );
    Colour col2 ( s,0,255,0 );
    Colour col3 ( s,255,255,0 );
    Colour col4 ( s,0,0,255 );
    Colour FG ( s, 0, 0, 0 );
    Colour BG ( s, 255, 255, 255 );

    Image* p_image;

    for ( int i=0; i<8; i++ )
    {
        for ( int j=0; j<8; j++ )
        {
            if ( ( i+j ) %2 )
            {
                p_image = square1;
            }
            else
            {
                p_image = square2;
            }

            int width = 30;
            int square_x = i*width;
            int square_y = j*width;

            if ( game_mode == ONE_PLAYER_BLACK  || ( game_mode == TWO_PLAYER && !whites_turn ) )
            {
                square_x = ( 7-i ) *width;
                square_y = ( 7-j ) *30;
            }

            p_image->Put ( s, square_x, square_y );

            if ( i == last_x && j == last_y )
            {
                s.Rect ( square_x, square_y, width, 30, col3 );
            }

            if ( i == board_x && j == board_y )
            {
                s.Rect ( square_x, square_y, width, 30, col1 );
            }

            if ( move.length() > 0 && i == select_x && j == select_y )
            {
                s.Rect ( square_x, square_y, width, 30, col2 );
            }

            if ( board.Get ( i,j ) != NULL )
            {
                board.Get ( i,j )->Draw ( s, square_x, square_y );
            }
        }
    }

    s.Rect ( 240,0, 80, 240, BG );

    Difference ( s, font, FG );

    if ( whites_turn )
    {
        font->PutString ( s, 242, 5, FG, Colour::NONE, "White's move" );
    }
    else
    {
        font->PutString ( s, 242, 5, FG, Colour::NONE, "Black's move" );
    }

    font->PutString ( s, 242, 25, FG, Colour::NONE, "Moves" );
    for ( int i=0; i < move_list_size && ( scroll+i ) < output.size(); i++ )
    {
        font->PutString ( s, 242, 45+12 * i, FG, Colour::NONE,
                          output[scroll+i] );
    }

    DisplayMessage ( s );
}



/**
 * Display a message
 * @param s
 */
void Chess2X::DisplayMessage ( Screen& s )
{
    Colour FG ( s, 0, 0, 0 );

    font->PutString ( s, 243, 190, FG,Colour::NONE, message[0] );
    font->PutString ( s, 243, 205, FG,Colour::NONE, message[1] );
    font->PutString ( s, 243, 220, FG,Colour::NONE, message[2] );
}



/**
 * DIsplay the current menu
 * @param s
 */
void Chess2X::DoMenu ( Screen& s )
{
    if ( square1 == NULL )
    {
        square1 = new Image ( s, "skins/" + *skin + "/square1.bmp" );
    }
    if ( square2 == NULL )
    {
        square2 = new Image ( s, "skins/" + *skin + "/square2.bmp" );
    }
    if ( menu_background == NULL )
    {
        menu_background = new Image ( s, "skins/" + *skin + "/menubg.bmp" );
    }

    Colour FG ( s, 0, 0, 0 );
    Colour BG ( s, 255, 255, 255 );
    Colour HL ( s, 255, 0, 0 );
    s.Rect ( 0, 0, s.GetWidth(), s.GetWidth(), FG );
    menu_background->Put ( s, 0, 0 );
    current_menu->Render ( s, *font, FG, BG, HL );

    string state;
    if ( saved )
    {
        state = "Saved";
    }
    else
    {
        state = "Unsaved";
    }

    in_game_menu->SetSubtitle( state );
}



/**
 * Main menu callback
 * @param data - Pointer to Chess2X
 */
void Chess2X::MainAction ( void* data )
{
    Chess2X* p_c = reinterpret_cast<Chess2X*> ( data );
    switch ( p_c->main_menu->GetOption() )
    {
    case 0:
        p_c->ResetGame();
        p_c->current_menu = p_c->start_menu;
        break;

    case 1:
        p_c->ResetGame();
        p_c->current_menu = p_c->main_menu;
        p_c->Load();
        break;

    case 2:
        p_c->current_menu = p_c->options_menu;
        break;

    }
}


/**
 * Start menu callback
 * @param data - Pointer to Chess2X
 */
void Chess2X::StartAction ( void* data )
{
    Chess2X* p_c = reinterpret_cast<Chess2X*> ( data );
    p_c->mode = GAME;
    switch ( p_c->start_menu->GetOption() )
    {
    case 0:
        p_c->game_mode = ONE_PLAYER_WHITE;
        p_c->Message ( "", "", "Select - Menu" );
        p_c->board_x = 4;
        p_c->board_y = 6;
        break;

    case 1:
        p_c->game_mode = ONE_PLAYER_BLACK;
        p_c->Message ( "Thinking","","X - Force move" );
        p_c->Send ( "white\ngo\n" );
        p_c->board_x = 4;
        p_c->board_y = 1;
        break;

    case 2:
        p_c->game_mode = TWO_PLAYER;
        p_c->Message ( "", "", "Select - Menu" );
        p_c->Send ( "force\n" );
        p_c->board_x = 4;
        p_c->board_y = 6;
        break;

    }

    p_c->last_x = p_c->board_x;
    p_c->last_y = p_c->board_y;
    p_c->history.push_back ( History ( p_c->current_move, p_c->board, "", true,
                                       p_c->last_x, p_c->last_y,
                                       p_c->board_x, p_c->board_y ) );
}

/**
 * In-game menu callback
 * @param data - Pointer to Chess2X
 */
void Chess2X::InGameAction ( void* data )
{
    Chess2X* p_c = reinterpret_cast<Chess2X*> ( data );
    switch ( p_c->in_game_menu->GetOption() )
    {
    case 0:
        p_c->mode = GAME;
        p_c->saved = false;
        break;

    case 1:
        p_c->mode = GAME;
        p_c->Send ( "hint\n" );
        break;

    case 2:
        p_c->mode = GAME;
        p_c->saved = false;
        if ( p_c->game_mode == TWO_PLAYER )
        {
            if ( p_c->history.size() > 1 )
            {
                // Back one move
                p_c->Restore();
                p_c->Send ( "undo\n" );
            }
        }
        else if ( p_c->history.size() > 2 )
        {
            // Back 2 moves
            p_c->history.pop_back();
            p_c->Restore();
            p_c->Send ( "remove\n" );
        }
        break;
		
	case 3:
        p_c->Save();
        break;

    case 4:
        p_c->mode = MENU;
        p_c->current_menu = p_c->main_menu;
        p_c->ResetGame();
        break;
    }
}

/**
 * Quit menu callback
 * @param data - Pointer to Chess2X
 */
void Chess2X::QuitAction ( void* data )
{
    Chess2X* p_c = reinterpret_cast<Chess2X*> ( data );
    if ( p_c->quit_menu->GetOption() == 1 )
    {
        if ( p_c->old_mode == MENU )
        {
            p_c->current_menu = p_c->main_menu;
        }
        if ( p_c->old_mode == GAME_MENU )
        {
            p_c->current_menu = p_c->in_game_menu;
        }
        p_c->mode = p_c->old_mode;
    }
    else
    {
        // Stop any thinking
        kill ( p_c->pid, SIGINT );

        p_c->Send ( "quit\n" );

        // Save set up
        p_c->config.SaveFile();
        exit ( 0 );
    }
}


/**
 * Promote menu callback
 * @param data - Pointer to Chess2X
 */
void Chess2X::PromoteAction ( void* data )
{
    Chess2X* p_c = reinterpret_cast<Chess2X*> ( data );

    string m ( p_c->move );
    switch ( p_c->promote_menu->GetOption() )
    {
    case 0:
        m += "q";
        break;

    case 1:
        m += "r";
        break;

    case 2:
        m += "n";
        break;

    case 3:
        m += "b";
        break;
    }

    p_c->mode = GAME;
    p_c->Send ( m+"\n" );
    p_c->move = "";

}

/**
 * Level menu callback
 * @param data - Pointer to Chess2X
 */
void Chess2X::LevelAction ( void* data )
{
    Chess2X* p_c = reinterpret_cast<Chess2X*> ( data );

    switch ( p_c->level_menu->GetOption() )
    {
    case 0:
        if ( p_c->moves_per_period < 100 )
        {
            p_c->moves_per_period += 10;
        }
        break;

    case 1:
        if ( p_c->moves_per_period > 10 )
        {
            p_c->moves_per_period -= 10;
        }
        break;
    }

    ostringstream oss;
    oss << "Level " << 11 - p_c->moves_per_period/10;
    p_c->level_menu->SetSubtitle ( oss.str() );
    p_c->config.SetValue ( "moves_per_period", p_c->moves_per_period );

}


/**
 * Options menu callback
 * @param data - Pointer to Chess2X
 */
void Chess2X::OptionsAction ( void* data )
{
    Chess2X* p_c = reinterpret_cast<Chess2X*> ( data );

    switch ( p_c->options_menu->GetOption() )
    {
    case 0:
        p_c->current_menu = p_c->level_menu;
        break;

    case 1:
        p_c->current_menu = p_c->skin_menu;
        break;
    }
}


/**
 * Skin menu callback
 * @param data - Pointer to Chess2X
 */
void Chess2X::SkinAction ( void* data )
{
    Chess2X* p_c = reinterpret_cast<Chess2X*> ( data );

    switch ( p_c->skin_menu->GetOption() )
    {
    case 0:
        p_c->skin++;
        if ( p_c->skin == p_c->skin_list.end() )  p_c->skin = p_c->skin_list.begin();
        p_c->SetSkin();
        break;

    case 1:
        if ( p_c->skin == p_c->skin_list.begin() )  p_c->skin = p_c->skin_list.end();
        p_c->skin--;
        p_c->SetSkin();
        break;
    }

    if ( p_c->skin_list.size() > 0 )
    {
        p_c->skin_menu->SetSubtitle ( * ( p_c->skin ) );
    }
}


/**
 * Save game
 */
void Chess2X::Save()
{
    ofstream save ( "save" );
    if ( save.good() )
    {
        save << game_mode << endl;
        for ( int i=0; i<output.size(); i++ )
        {
            save << output[i] << endl;
        }
        save << "end" << endl;
    }
    save.close();
    sync();
    saved = true;
}



/**
 * Load game
 */
void Chess2X::Load()
{
    ifstream save ( "save" );
    if ( save.good() )
    {
        Send ( "force\n" );
        int gm = 0;
        save >> gm;
        game_mode = static_cast<GameMode> ( gm );
        output.clear();
        scroll = 0;
        current_move = 1;
        whites_turn = true;
        loaded = true;
        move_queue.clear();
        bool end = false;

        while ( !end )
        {
            string move, white, black;
            save >> move;
            end = ( move == "end" );
            if ( !end )
            {
                save >> white >> black;

                if ( white != "" )
                {
                    move_queue.push_back ( white+"\n" );
                    whites_turn = !whites_turn;
                }

                if ( black != "end" )
                {
                    move_queue.push_back ( black+"\n" );
                    whites_turn = !whites_turn;
                }
                else
                {
                    end = true;
                }
            }
        }
        save.close();

        if ( !move_queue.empty() )
        {
            // Send first move
            Send ( move_queue.front() );
            move_queue.pop_front();
        }

        switch ( game_mode )
        {
        case ONE_PLAYER_WHITE:
            board_x = 4;
            board_y = 6;
            break;

        case ONE_PLAYER_BLACK:
            board_x = 4;
            board_y = 1;
            break;

        case TWO_PLAYER:
            board_x = 4;
            if ( whites_turn )
            {
                board_y = 6;
            }
            else
            {
                board_y = 1;
            }
            break;
        }

        last_x = board_x;
        last_y = board_y;

        mode = GAME;
		
		Message("", "", "Select - Menu");
    }
    else
    {
        current_menu = main_menu;
    }
}


/**
 * Set game to initial values
 */
void Chess2X::ResetGame()
{
    Send ( "new\nrandom\nprotover 2\nhard\n" );
    ostringstream oss;
    oss << "level " << moves_per_period << " 5 0\n";
    Send ( oss.str() );
    oss.str ( "" );
    oss << "hashsize " << hash_size << "\n";
    Send ( oss.str() );
    board.Setup ( "skins/" + *skin );
    output.clear();
    scroll = 0;
    whites_turn = true;
    loaded = false;
    current_move = 1;
    saved = false;
    win = false;
    history.clear();
}



/**
 * Check for illegal moves
 * @param line - Chess engine output line
 * @return True if move is illegal
 */
bool Chess2X::Illegal ( const std::string& line )
{
    const string ILLEGAL ( "Illegal move" );
    string::size_type pos = line.find ( ILLEGAL );
    if ( pos == string::npos )
    {
        return false;
    }

    if ( line[pos + ILLEGAL.length() ] == ':' )
    {
        pos++;
    }
    Message ( "Illegal move", line.substr ( pos + ILLEGAL.length() ) );
}


/**
 * Check for Hints
 * @param line - Chess engine output line
 * @return True if hint
 */
bool Chess2X::Hint ( const std::string& line )
{
    const string HINT ( "Hint" );
    string::size_type pos = line.find ( HINT );
    if ( pos == string::npos )
    {
        return false;
    }

    if ( line[pos + HINT.length() ] == ':' )
    {
        pos++;
    }

    string hint = line.substr ( pos + HINT.length() );
    last_x = hint[hint.length()-2] - 'a';
    last_y = 8- ( hint[hint.length()-1] - '0' );
    Message ( "Hint", hint, "Select - Menu" );
}


/**
 * Make a list a skins from the directories in the skins directory
 */
void Chess2X::GetSkins()
{
    skin_list.clear();
    DIR *d = opendir ( "skins" );
    if ( d != NULL )
    {
        dirent *ent = NULL;
        while ( ( ent = readdir ( d ) ) != NULL )
        {
            if ( ent->d_type == DT_DIR && ent->d_name[0] != '.' )
            {
                skin_list.push_back ( ent->d_name );
            }
        }
        closedir ( d );
    }

    sort ( skin_list.begin(), skin_list.end() );

    skin = skin_list.begin();

    string s;
    if ( config.GetValue ( "skin", s ) )
    {
        vector<string>::iterator it = find ( skin_list.begin(), skin_list.end(), s );
        if ( it != skin_list.end() )
        {
            skin = it;
        }
    }
    else
    {
        config.SetValue ( "skin", *skin );
    }
}



/**
 * Reset all skin images
 */
void Chess2X::SetSkin()
{
    if ( square1 != NULL ) delete square1;
    if ( square2 != NULL ) delete square2;
    if ( menu_background != NULL ) delete menu_background;

    square1 = NULL;
    square2 = NULL;
    menu_background = NULL;

    config.SetValue ( "skin", *skin );
}


/**
 * Read volume from configuration file
 */
void Chess2X::GetVolume()
{
    int v = 64;
    if ( config.GetValue ( "fxvolume", v ) )
    {
        move_sound->SetVolume ( v );
        take_sound->SetVolume ( v );
    }
    else
    {
        config.SetValue ( "fxvolume", move_sound->GetVolume() );
    }

    v = 64;
    if ( config.GetValue ( "musicvolume", v ) )
    {
        player.SetVolume ( v );
    }
    else
    {
        config.SetValue ( "musicvolume", player.GetVolume() );
    }

}


/**
 * Move cursor up the board
 */
void Chess2X::UpSquare()
{
    if ( game_mode == ONE_PLAYER_BLACK || ( game_mode == TWO_PLAYER && !whites_turn ) )
    {
        board_y = ( board_y+1 ) %8;
    }
    else
    {
        board_y = ( board_y+7 ) %8;
    }
}


/**
 * Move cursor down the board
 */
void Chess2X::DownSquare()
{
    if ( game_mode == ONE_PLAYER_BLACK || ( game_mode == TWO_PLAYER && !whites_turn ) )
    {
        board_y = ( board_y+7 ) %8;
    }
    else
    {
        board_y = ( board_y+1 ) %8;
    }
}


/**
 * Move cursor left on the board
 */
void Chess2X::LeftSquare()
{
    if ( game_mode == ONE_PLAYER_BLACK || ( game_mode == TWO_PLAYER && !whites_turn ) )
    {
        board_x = ( board_x+1 ) %8;
    }
    else
    {
        board_x = ( board_x+7 ) %8;
    }
}


/**
 * Move cursor right on the board
 */
void Chess2X::RightSquare()
{
    if ( game_mode == ONE_PLAYER_BLACK || ( game_mode == TWO_PLAYER && !whites_turn ) )
    {
        board_x = ( board_x+7 ) %8;
    }
    else
    {
        board_x = ( board_x+1 ) %8;
    }
}



/**
 * Restore position from history
 */
void Chess2X::Restore()
{
    // Current move is last move
    current_move = history.back().GetMoveNo();

    // Discard last move
    history.pop_back();

    // Restore game state
    board = history.back().GetBoard();
    whites_turn = history.back().GetWhitesTurn();
    history.back().GetLastPosition ( last_x, last_y );
    history.back().GetThisPosition ( board_x, board_y );
    if ( history.back().GetMove() == "" )
    {
        Message ( "", "", "Select - Menu" );
    }
    else
    {
        Message ( "Last move ", history.back().GetMove(), "Select - Menu" );
    }
    move="";

    // Rebuild move list
    output.clear();
    for ( list<History>::iterator it = history.begin();
            it != history.end(); it++ )
    {
        int n = it->GetMoveNo();
        string m = it->GetMove();
        if ( m != "" )
        {
            cout << "History: " << n << " " << m
                 << " list size = " << output.size() << endl;
            if ( output.size() < n )
            {
                ostringstream oss;
                oss << setw ( 2 ) << n << " " << m << " ";
                output.insert ( output.end(), oss.str() );
            }
            else
            {
                output[n-1] += m;
            }
        }
    }

    scroll = 0;

    if ( current_move > move_list_size )
    {
        scroll = current_move - move_list_size;
    }
}


/**
 * Read the sound files from skin or use defaults
 */
void Chess2X::GetSounds()
{
    move_sound = GetSound ( "move_sound" );
    take_sound = GetSound ( "take_sound" );
}


/**
 * Read the sound file from the skin or use the default
 * @param fn - File name
 * @return Sound pointer
 */
Sound* Chess2X::GetSound ( const std::string& fn )
{
    string mp = *skin + "/" + fn;
    if ( access ( mp.c_str(), R_OK ) != 0 )
    {
        mp = fn;
    }

    cout << "Sound file: " << mp << endl;
    return new Sound ( mp );
}



/**
 * Show piece differences
 * @param s - Screen
 * @param f - Font
 * @param c - Text colour
*/
void Chess2X::Difference ( Screen &s, Font* f, const Colour& c )
{
    Diff ( PAWN, s, f, c, 240, 95 );
    Diff ( BISHOP, s, f, c, 280, 95 );
    Diff ( KNIGHT, s, f, c, 240, 125 );
    Diff ( ROOK, s, f, c, 280, 125 );
    Diff ( QUEEN, s, f, c, 240, 155 );
}
/**
 * Show piece difference
 * @param t - Type
 * @param s - Screen
 * @param f - Font
 * @param c - Text colour
 * @param x
 * @param y
*/
void Chess2X::Diff ( PieceType t, Screen &s, Font* f, const Colour& c, int x, int y )
{
    int n=0;
    Piece *p = board.Taken ( t, n );
    if ( p != NULL )
    {
        ostringstream oss;
        oss << n;
        p->Draw ( s, x, y );
        f->PutString ( s, x + 32, y + 10, c, Colour::NONE, oss.str() );
    }
}



/**
 * Music player
 * @param s - Screen
 */
void Chess2X::Player ( Screen& s )
{
    menu_background->Put ( s, 0, 0 );
    Colour FG ( s, 0, 0, 0 );
    Colour BG ( s, 255, 255, 255 );
    Colour HL ( s, 255, 0, 0 );
    player.Render ( s, *font, FG, BG, HL );

    font->PutString ( s, 10, 200, FG, BG,
                      "A - Exit player  B - Start/Stop" );
    font->PutString ( s, 10, 220, FG, BG, "Y - Add file  X - Remove file" );

}

/**
 * File Selctor
 * @param s - Screen
 */
void Chess2X::FileSelect ( Screen& s )
{
    menu_background->Put ( s, 0, 0 );
    Colour FG ( s, 0, 0, 0 );
    Colour BG ( s, 255, 255, 255 );
    Colour HL ( s, 255, 0, 0 );
    selector.Render ( s, *font, FG, BG, HL );

    font->PutString ( s, 10, 220, FG, BG,
                      "Y - Back to player  B - Select X - Up a level" );
}


void Chess2X::AudioFileSelected ( void* data, std::string full, std::string fn )
{
    Chess2X* p_c = reinterpret_cast<Chess2X*> ( data );
    if ( p_c != NULL )
    {
        if ( p_c->player.AddFile ( full ) )
        {
            p_c->selector.SetMessage ( "Added " + fn );
        }
        else
        {
            p_c->selector.SetMessage ( fn + " is not a supported audio file" );
        }
    }
}
