/***************************************************************************
 *   Copyright (C) 2006 by Dave Parker                                     *
 *   drp@ezeedsl.co.uk                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef CHESS2X_H
#define CHESS2X_H
#include <vector>
#include <string>
#include <list>
#include "chesscommon.h"
#include "impl.h"
#include "board.h"
#include "chessconfig.h"
#include "menu.h"
#include "piece.h"
#include "sound.h"
#include "history.h"
#include "audioplayer.h"
#include "fileselector.h"
#include <time.h>

class Font;
class Image;

/**
 * Class Chess2X - Main application
 *
 */
class Chess2X : public Impl
{
public:

    // Initialisation
    virtual void Initialise ( Screen& s );

    // Render screen
    virtual void Render ( Screen& s );

    // Receive xgrep output
    virtual void NonRender ( const Input& input );

    // Handle button presses
    virtual void ButtonChange ( const Input& input, Input::ButtonType b );
    void Message ( const std::string& m, const std::string& m2 = "", const std::string& m3 = "" );
    void Parse ( const std::string line );
    bool Win ( const std::string& line );
    void DisplayMessage ( Screen& s );
    void DoMenu ( Screen& s );
    static void MainAction ( void* data );
    static void StartAction ( void* data );
    static void InGameAction ( void* data );
    static void QuitAction ( void* data );
    static void PromoteAction ( void* data );
    static void LevelAction ( void* data );
    static void OptionsAction ( void* data );
    static void SkinAction ( void* data );

    void ResetGame();
    bool Hint ( const std::string& line ); // Detect hint
    bool Illegal ( const std::string& line ); // Detect illegal move
    void GetSkins();
    void SetSkin();
    void GetVolume();
    void UpSquare();   // Move cursor up a square
    void DownSquare(); // Move cursor down a square
    void LeftSquare(); // Move cursor left a square
    void RightSquare(); // Move cursor right a square
    void Difference ( Screen &s, Font* f, const Colour& c ); // Show piece differences
    void Diff ( PieceType t, Screen &s, Font* f, const Colour& c, int x, int y ); // Show piece difference
    void Player ( Screen& s );
    void FileSelect ( Screen& s );
    static void AudioFileSelected ( void* data, std::string full, std::string fn );

    enum DisplayMode
    {
        MENU,
        GAME_MENU,
        GAME,
        AUDIO_PLAYER,
        FILE_SELECT
    };

    enum GameMode
    {
        ONE_PLAYER_WHITE,
        ONE_PLAYER_BLACK,
        TWO_PLAYER
    };

private:
    Font*  font;              // Font
    Image* menu_background;   // Background menu image
    Menu*  main_menu;	      // Main menu
    Menu*  start_menu;	      // Start menu
    Menu*  in_game_menu;      // In-game menu
    Menu*  quit_menu;	      // Quit menu
    Menu*  promote_menu;      // Promote menu
    Menu*  level_menu;        // Level menu
    Menu*  options_menu;      // Options menu
    Menu*  skin_menu;         // Skin selection menu

    Menu*  current_menu;      // Menu currently displayed

    Board board;              // Chess board
    int board_x;              // cursor x position
    int board_y;              // cursory position
    int select_x;             // selected piece x position
    int select_y;             // selected piece y position
    std::string move;         // Move
    Image* square1;           // Image for square
    Image* square2;           // Image for square
    Sound* move_sound;        // Sound for move
    Sound* take_sound;        // Sound for take

    std::list<History> history; // Move history

    std::vector<std::string> output; // Output moves
    int inpipefd[2];          // Pipe to gnuchess
    int outpipefd[2];         // Pipe from gnuchess
    pid_t pid;                // Process id of gnuchess
    int scroll;               // Output line at top of screen

    void Run ( void );        // Run engine
    void Game ( Screen& s );  // Display Game
    void Send ( const std::string& s ); // Send command to engine
    std::string message[3];   // Message 3 lines
    bool whites_turn;         // True if white's turn

protected:
    int move_list_size;       // Size for move list displayed
    DisplayMode mode;         // Mode - Game/Menu
    DisplayMode old_mode;     // Previous mode
    GameMode game_mode;       // Game mode 1/2 player etc
    int current_move;         // Current move number

    std::vector<std::string> skin_list;      // List of skins
    std::vector<std::string>::iterator skin; // Selected skin

    std::list<std::string> move_queue; // Move queue for loading
    bool loaded; // True if game has just been loaded
    bool saved;  // True if game is saved
    bool win;    // True if game is finished
    int moves_per_period;  // Level
    int last_x;	 // Last move X
    int last_y;  // Last move Y
    int hash_size;  // Hash size

    ChessConfig config;  // Saved configuration
    
    std::vector<Input::ButtonType> repeats;  // Repeated keys

private:
    void Save(); // Save game
    void Load(); // Load game

protected:
    void Restore();    		// Restore position from history
    void GetSounds();  		// Get sounds
    Sound* GetSound ( const std::string& fn );  // Get sound
    AudioPlayer player; 	// Audio player
    FileSelector selector; 	// File selector
};
#endif //CHESS2X_H

