#include "colour.h"
#include "piece.h"
#include "image.h"
#include "screen.h"
#include <iostream>
#include <sstream>

using namespace std;


/**
 * Construct a piece
 * @param fn - Image file
 * @param t - Type
 * @param s - Side
 * @return
 */
Piece::Piece ( const string& fn, PieceType t, Side s ) : file_name ( fn ), type ( t ), side ( s ), image ( NULL )
{
}


/**
 * Copy constructor
 * @param p - Source piece
 * @return
 */
Piece::Piece ( const Piece& p ) : file_name ( p.file_name ), type ( p.type ), side ( p.side ), image ( NULL )
{
}


/**
 * Destructor
 * @return
 */
Piece::~Piece()
{
}

/**
 * Draw a piece at a given position
 * @param s - Destination Surface
 * @param x
 * @param y
 */
void Piece::Draw ( Screen& s, int x, int y )
{
    if ( image == NULL )
    {
        ostringstream oss;
        oss << file_name << side << ".bmp";
        image = new Image ( s, oss.str() );
    }

    image->PutBlended ( s, x, y, image->GetPixel ( 0,0 ) );
}



/**
 * Move piece
 * @param board
 * @param x1
 * @param y1
 * @param x2
 * @param y2
 */
void Piece::Move ( Board& board, int x1, int y1, int x2, int y2 )
{
    if ( board.Get ( x2,y2 ) != NULL )
    {
        board.Take ( x2,y2 );
    }
    board.Set ( x1, y1, NULL );
    board.Set ( x2, y2, this );
}


/**
 * Check for promotion of a piece - default is always false
 * @param x1
 * @param y1
 * @param x2
 * @param y2
 * @return true if the piece can be promoted
 */
bool Piece::CheckPromote ( int x1, int y1, int x2, int y2 )
{
    return false;
}


/**
 * Set the image file for a piece
 * @param fn
 */
void Piece::SetImage ( const std::string& fn )
{
    if ( image != NULL )
    {
        delete image;
        image = NULL;
    }

    file_name = fn;
}


/**
 * Construct a rook
 * @param s - Side
 * @param i - Image file
 * @return
 */
Rook::Rook ( Side s, const std::string &i ) : Piece ( i, ROOK, s )
{
}


/**
 * Construct a knight
 * @param s - Side
 * @param i - Image file
 * @return
 */
Knight::Knight ( Side s, const std::string &i ) : Piece ( i, KNIGHT, s )
{
}


/**
 * Construct a bishop
 * @param s - Side
 * @param i - Image file
 * @return
 */
Bishop::Bishop ( Side s, const std::string &i ) : Piece ( i, BISHOP, s )
{
}


/**
 * Construct a queen
 * @param s - Side
 * @param i - Image file
 * @return
 */
Queen::Queen ( Side s, const std::string &i ) : Piece ( i, QUEEN, s )
{
}


/**
 * Construct a king
 * @param s - Side
 * @param i - Image file
 * @return
 */
King::King ( Side s, const std::string &i ) : Piece ( i, KING, s )
{
}


/**
 * Move king checking for castling
 * @param board
 * @param x1
 * @param y1
 * @param x2
 * @param y2
 */
void King::Move ( Board& board, int x1, int y1, int x2, int y2 )
{
    Piece::Move ( board, x1, y1, x2, y2 );
    if ( x1 == 4 && ( y1 == 0 || y1 ==7 ) && y1 == y2 )
    {
        if ( x2 == 6 )
        {
            // Castle king's side
            board.Set ( 5, y1, board.Get ( 7,y1 ) );
            board.Set ( 7, y1, NULL );
        }
        else if ( x2 == 2 )
        {
            // Castle queen's side
            board.Set ( 3, y1, board.Get ( 0,y1 ) );
            board.Set ( 0, y1, NULL );
        }
    }
}

/**
 * Construct a pawn
 * @param s - Side
 * @param i - Image file
 * @return
 */
Pawn::Pawn ( Side s, const std::string &i ) : Piece ( i, PAWN, s )
{
}


/**
 * Move pawn checking for en passent
 * @param board
 * @param x1
 * @param y1
 * @param x2
 * @param y2
 */
void Pawn::Move ( Board& board, int x1, int y1, int x2, int y2 )
{
    if ( ( abs ( x2 - x1 ) == 1 ) && board.Get ( x2, y2 ) == NULL &&
            ( ( y1 == 3 && y2 == 2 ) || ( y1 == 4 && y2 == 5 ) ) )
    {
        cout << "en passant" << endl;
        // en passant
        board.Take ( x2, y1 );
    }
    Piece::Move ( board, x1, y1, x2, y2 );
}


/**
 * Check for pawn promotion
 * @param x1
 * @param y1
 * @param x2
 * @param y2
 * @return
 */
bool Pawn::CheckPromote ( int x1, int y1, int x2, int y2 )
{

    return ( (side == WHITE && y2 == 0 && y1 == 1) ||
             (side == BLACK && y2 == 7 && y1 == 6 ));
}




